#include <vector>
#include <iostream>
#include <fstream>
#include "Shader.h"

inline const std::string getGLSLErrorString(const unsigned i) {
	const char* const aGLSLErrorString[] = {
        "(e0000) GLSL not enabled",
        "(e0001) not a valid program object",
        "(e0002) not a valid object",
        "(e0003) out of memory",
        "(e0004) unknown compiler error"};
	return aGLSLErrorString[i];
}

Shader::Shader(): compiler_log(0), is_compiled(false),
	ProgramObject(0), ShaderSource(0), _memalloc(false) {
}

Shader::~Shader() {
   if (compiler_log != 0) 
	   free(compiler_log);

   if (ShaderSource != 0) {
        if (_memalloc)
            delete[] ShaderSource;  // free ASCII Source
   }
   
   if (is_compiled) { 
        glDeleteObjectARB(ProgramObject);
        CHECK_GL_ERROR();
   }
}

unsigned long getFileSize(std::ifstream& file) {
    using namespace std;
	if(!file.good()) return 0;
    
    file.tellg();
    file.seekg(0,ios::end);
    unsigned long len = file.tellg();
    file.seekg(ios::beg);
    
    return len;
}

int Shader::load(const std::string& filename) {
    using namespace std;
	ifstream file;
	file.open(filename.c_str(), ios::in);
    if(!file) return -1;
    
    unsigned long len = getFileSize(file);
    
    if (len==0) return -2;   // "Empty File" 
    
	if (ShaderSource!=0) {
		// there is already a source loaded, free it!
        if (_memalloc)
	        delete[] ShaderSource;
    }
    
    ShaderSource = (GLubyte*) new char[len+1];
    if (ShaderSource == 0) return -3;   // can't reserve memory
    _memalloc = true;
    
    
    ShaderSource[len] = 0;  // len isn't always strlen cause some characters are stripped in ascii read...
                            // it is important to 0-terminate the real length later, len is just max possible value...
   
    unsigned int i=0;
    while (file.good())
    {
        ShaderSource[i++] = file.get();       // get character from file
        if (i>len) i=len;   // coding guidelines...
    }
    
    ShaderSource[i] = 0;  // 0 terminate it.
    
    file.close();
      
	return 0;
}

void Shader::loadFromMemory(const char* program) {
    if (ShaderSource!=0)    // there is already a source loaded, free it!
    {
        if (_memalloc)
        delete[] ShaderSource;
    }
   _memalloc = false;
   ShaderSource = (GLubyte*) program;
      
}

const std::string Shader::getCompilerLog(void) {    
	int blen = 0;	
	int slen = 0;	


	if (ProgramObject == 0) 
		return getGLSLErrorString(1);

	glGetObjectParameterivARB(ProgramObject, GL_OBJECT_INFO_LOG_LENGTH_ARB , &blen);
	CHECK_GL_ERROR();

	if (blen > 1) {
		if (compiler_log!=0) {   
			free(compiler_log);
			compiler_log =0;
		}
		if ((compiler_log = (GLcharARB*)malloc(blen)) == NULL) {
			printf("ERROR: Could not allocate compiler_log buffer\n");
			return getGLSLErrorString(3);
		}

		glGetInfoLogARB(ProgramObject, blen, &slen, compiler_log);
		CHECK_GL_ERROR();
	}
	if (compiler_log != 0)
		return (char*) compiler_log;    

	return getGLSLErrorString(4);
}

bool Shader::compile(void)
{

	is_compiled = false;

	int compiled = 0;

	if (ShaderSource == 0) 
		return false;

	GLint	length = (GLint) strlen((const char*)ShaderSource);
	glShaderSourceARB(ProgramObject, 1, (const GLcharARB **)&ShaderSource, &length);
	CHECK_GL_ERROR();

	glCompileShaderARB(ProgramObject); 
	CHECK_GL_ERROR();
	glGetObjectParameterivARB(ProgramObject, GL_OBJECT_COMPILE_STATUS_ARB, &compiled);
	CHECK_GL_ERROR();

	if (compiled) 
		is_compiled = true;
	 
	return is_compiled;
}

VertexShader::VertexShader() {
	ProgramObject = glCreateShaderObjectARB(GL_VERTEX_SHADER_ARB);
	CHECK_GL_ERROR();
}

VertexShader::~VertexShader() {
}

FragmentShader::FragmentShader() {
	ProgramObject = glCreateShaderObjectARB(GL_FRAGMENT_SHADER_ARB); 
	CHECK_GL_ERROR();
}

FragmentShader::~FragmentShader() {
}

